#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          14-08-2018
//      Created for Project :   FLAME
"""

# Python standard libs
import logging
from hashlib import sha256

# PIP installed libs
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_defaults, view_config
from yaml import load, YAMLError
from toscaparser.tosca_template import ToscaTemplate
from requests import post

# CLMC-service imports
from clmcservice.alertsapi.utilities import adjust_tosca_definitions_import, TICKScriptTemplateFiller, fill_http_post_handler_vars, get_resource_spec_policy_triggers, get_alert_spec_policy_triggers
from clmcservice.alertsapi.alerts_specification_schema import COMPARISON_OPERATORS, SFEMC, validate_clmc_alerts_specification

# initialise logger
log = logging.getLogger('service_logger')


@view_defaults(renderer='json')
class AlertsConfigurationAPI(object):
    """
    A class-based view for configuring alerts within CLMC.
    """

    STREAM_PERIOD_LIMIT = 60  # if alert period is <= 60 seconds, then a stream template is used, otherwise use batch

    DUAL_VERSION_TEMPLATES = {"threshold"}  # this set defines all template types that are written in two versions (stream and batch)

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='alerts_configuration', request_method='GET')
    def get_alerts_hash(self):
        """
        Retrieves hash value for alerts task, topic and handlers based on sfc, sfci, policy and trigger IDs
        """

        for param in ("sfc", "sfci", "policy", "trigger"):
            if param not in self.request.params:
                raise HTTPBadRequest("Missing url query string parameter {0}.".format(param))

        sfc = self.request.params["sfc"]
        sfci = self.request.params["sfci"]
        policy = self.request.params["policy"]
        trigger = self.request.params["trigger"]

        # generate topic and alert identifiers
        topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfci, policy, trigger)
        topic_id = self.get_hash(topic_id)
        alert_id = topic_id

        return {
            "task_identifier": alert_id,
            "task_api_endpoint": "/kapacitor/v1/tasks/{0}".format(alert_id),
            "topic_identifier": topic_id,
            "topic_api_endpoint": "/kapacitor/v1/alerts/topics/{0}".format(topic_id),
            "topic_handlers_api_endpoint": "/kapacitor/v1/alerts/topics/{0}/handlers".format(topic_id)
        }

    @view_config(route_name='alerts_configuration', request_method='POST')
    def post_alerts_specification(self):
        """
        The view for receiving and configuring alerts based on the TOSCA alerts specification document. This endpoint must also receive the TOSCA resources specification document for validation.

        :raises HTTPBadRequest: if the request doesn't contain a (YAML) file input referenced as alert-spec representing the TOSCA Alerts Specification
        """

        kapacitor_host, kapacitor_port = self.request.registry.settings['kapacitor_host'], self.request.registry.settings['kapacitor_port']

        alert_spec_reference = self.request.POST.get('alert-spec')
        resource_spec_reference = self.request.POST.get('resource-spec')

        # check that the resource specification file was sent
        if not hasattr(resource_spec_reference, "file") or not hasattr(resource_spec_reference, "filename"):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'resource-spec' representing the TOSCA Resource Specification.")

        try:
            resource_spec_sfc, resource_spec_sfc_i, resource_spec_policy_triggers = get_resource_spec_policy_triggers(resource_spec_reference)
        except Exception as e:
            log.error("Couldn't extract resource specification event IDs due to error: {0}".format(e))
            raise HTTPBadRequest("Couldn't extract resource specification event IDs - invalid TOSCA resource specification.")

        # check that the alerts specification file was sent
        if not hasattr(alert_spec_reference, "file") or not hasattr(alert_spec_reference, "filename"):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'alert-spec' representing the TOSCA Alert Specification.")

        # extract alert specification file and filename
        alerts_input_filename = alert_spec_reference.filename
        alerts_input_file = alert_spec_reference.file

        if not alerts_input_filename.lower().endswith('.yaml'):
            raise HTTPBadRequest("Request to this API endpoint must include a (YAML) file input referenced as 'alert-spec' representing the TOSCA Alerts Specification.")

        # parse the alerts specification file
        try:
            alerts_yaml_content = load(alerts_input_file)
            adjust_tosca_definitions_import(alerts_yaml_content)
        except YAMLError as err:
            log.error("Couldn't parse user request file {0} to yaml format due to error: {1}".format(alerts_input_filename, err))
            log.error("Invalid content is: {0}".format(alerts_input_file.read()))
            raise HTTPBadRequest("Request alert specification file could not be parsed as valid YAML document.")

        try:
            tosca_tpl = ToscaTemplate(yaml_dict_tpl=alerts_yaml_content)
        except Exception as e:
            log.error(e)
            raise HTTPBadRequest("Request alert specification file could not be parsed as a valid TOSCA document.")

        valid_alert_spec = validate_clmc_alerts_specification(tosca_tpl.tpl)
        if not valid_alert_spec:
            raise HTTPBadRequest("Request alert specification file could not be validated as a CLMC TOSCA alerts specification document.")

        alert_spec_policy_triggers = get_alert_spec_policy_triggers(tosca_tpl)
        # TODO next release - uncomment
        # sfc, sfc_instance = tosca_tpl.tpl["metadata"]["sfc"], tosca_tpl.tpl["metadata"]["sfci"]
        sfc = tosca_tpl.tpl["metadata"]["servicefunctionchain"]
        sfc_instance = "{0}_1".format(sfc)

        # do validation between the two TOSCA documents
        self._compare_alert_and_resource_spec(sfc, sfc_instance, alert_spec_policy_triggers, resource_spec_sfc, resource_spec_sfc_i, resource_spec_policy_triggers)

        db = sfc  # database per service function chain, named after the service function chain ID

        # two lists to keep track of any errors while interacting with the Kapacitor HTTP API
        alert_tasks_errors = []
        alert_handlers_errors = []

        # iterate through every policy and extract all triggers of the given policy
        self._config_kapacitor_alerts(tosca_tpl, sfc, sfc_instance, db, kapacitor_host, kapacitor_port, resource_spec_policy_triggers, alert_tasks_errors, alert_handlers_errors)

        return_msg = {"msg": "Alerts specification has been successfully validated and forwarded to Kapacitor", "service_function_chain_id": sfc,
                      "service_function_chain_instance_id": sfc_instance}

        if len(alert_tasks_errors) > 0:
            return_msg["triggers_specification_errors"] = alert_tasks_errors
            self.request.response.status = 400

        if len(alert_handlers_errors) > 0:
            return_msg["triggers_action_errors"] = alert_handlers_errors
            self.request.response.status = 400

        return return_msg

    def _compare_alert_and_resource_spec(self, alert_spec_sfc, alert_spec_sfc_instance, alert_spec_policy_triggers, resource_spec_sfc, resource_spec_sfc_instance, resource_spec_policy_triggers):
        """
        Compares the extracted values from the resource spec against the values from the alerts spec - validation that they refer to the same things,

        :param alert_spec_sfc: sfc from alert spec
        :param alert_spec_sfc_instance: sfc instance from alert spec
        :param alert_spec_policy_triggers: policy/trigger IDs from alert spec
        :param resource_spec_sfc: sfc from resource spec
        :param resource_spec_sfc_instance: sfc instance from resource spec
        :param resource_spec_policy_triggers: policy/trigger IDs from resource spec

        :raises: HTTP Bad Request if the two specifications are inconsistent
        """

        if alert_spec_sfc != resource_spec_sfc:
            raise HTTPBadRequest("Different service function chain ID used in the alert and resource specification documents: {0} != {1}".format(alert_spec_sfc, resource_spec_sfc))

        if alert_spec_sfc_instance != resource_spec_sfc_instance:
            raise HTTPBadRequest("Different service function chain instance ID used in the alert and resource specification documents: {0} != {1}".format(alert_spec_sfc_instance, resource_spec_sfc_instance))

        alert_spec_triggers_set = set(alert_spec_policy_triggers)
        resource_spec_triggers_set = set(resource_spec_policy_triggers)

        # check for triggers defined in the resource specification but missing in the alerts specification
        missing_policy_triggers = [policy_trigger_id.replace("\n", " : ") for policy_trigger_id in resource_spec_policy_triggers if policy_trigger_id not in alert_spec_triggers_set]
        if len(missing_policy_triggers) > 0:
            raise HTTPBadRequest("Couldn't match the following policy triggers from the resource specification with triggers defined in the alerts specification: {0}".format(missing_policy_triggers))

        # check for triggers defined in the alerts specification and posting to flame_sfemc, but missing in the resource specification
        missing_policy_triggers = [policy_trigger_id.replace("\n", " : ") for policy_trigger_id in alert_spec_policy_triggers if policy_trigger_id not in resource_spec_triggers_set]
        if len(missing_policy_triggers) > 0:
            raise HTTPBadRequest("Couldn't match the following policy triggers from the alerts specification with triggers defined in the resource specification: {0}".format(missing_policy_triggers))

    def _config_kapacitor_alerts(self, tosca_tpl, sfc, sfc_instance, db, kapacitor_host, kapacitor_port, resource_spec_policy_triggers, alert_tasks_errors, alert_handlers_errors):
        """
        Configures the alerts task and alert handlers within Kapacitor.

        :param tosca_tpl: the parsed Tosca template object
        :param sfc: sfc ID
        :param sfc_instance: sfc instance ID
        :param db: Influx database ID
        :param kapacitor_host: default host is localhost (CLMC service running on the same machine as Kapacitor)
        :param kapacitor_port: default value to use is 9092
        :param resource_spec_policy_triggers: the extracted policy-trigger strings from the resource specification
        :param alert_tasks_errors: the list for tracking errors while interacting with Kapacitor tasks
        :param alert_handlers_errors: the list for tracking errors while interacting with Kapacitor alert handlers

        :return: the list of successfully registered event identifiers
        """

        for policy in tosca_tpl.policies:
            for trigger in policy.triggers:
                event_id = trigger.name
                policy_id = policy.name
                resource_spec_trigger_id = resource_spec_policy_triggers.get("{0}\n{1}".format(policy_id, event_id))

                event_type = trigger.trigger_tpl["event_type"]
                template_id = "{0}-template".format(event_type)
                measurement, field = trigger.trigger_tpl["metric"].split(".")

                condition = trigger.trigger_tpl["condition"]
                critical_value = float(condition["threshold"])
                alert_period_integer = condition["granularity"]
                alert_period = "{0}s".format(alert_period_integer)
                influx_function = condition.get("aggregation_method", "mean")  # if not specified, use "mean"

                # check for tag filtering
                where_clause = None
                if "resource_type" in trigger.trigger_tpl["condition"]:
                    tags = condition["resource_type"]
                    # make sure alert tasks are executing with queries for the given sfc and sfc instance - automatically add those tags using the metadata values
                    tags["flame_sfc"] = sfc
                    tags["flame_sfci"] = sfc_instance

                    # build up the where clause from the tags dictionary
                    where_filters_list = map(lambda tag_name: '"{0}"=\'{1}\''.format(tag_name, tags[tag_name]), tags)
                    where_clause = " AND ".join(where_filters_list)

                comparison_operator = COMPARISON_OPERATORS[condition.get("comparison_operator", "gte")]  # if not specified, use "gte" (>=)

                # generate topic and alert identifiers
                topic_id = "{0}\n{1}\n{2}\n{3}".format(sfc, sfc_instance, policy_id, event_id)  # scoped per service function chain instance (no two sfc instances report to the same topic)
                topic_id = self.get_hash(topic_id)
                alert_id = topic_id

                # check whether the template needs to be a stream or a batch
                if event_type in self.DUAL_VERSION_TEMPLATES:
                    if alert_period_integer <= self.STREAM_PERIOD_LIMIT:
                        template_id = "{0}-stream-template".format(event_type)
                        event_type = "{0}_stream".format(event_type)
                    else:
                        template_id = "{0}-batch-template".format(event_type)
                        event_type = "{0}_batch".format(event_type)

                # built up the template vars dictionary depending on the event type (threshold, relative, etc.)
                # all extracted properties from the trigger are passed, the TICKScriptTemplateFiller entry point then forwards those to the appropriate function for template filling
                template_vars = TICKScriptTemplateFiller.fill_template_vars(event_type, sfc=sfc, sfci=sfc_instance, policy=policy_id, db=db, measurement=measurement,
                                                                            field=field, influx_function=influx_function, critical_value=critical_value, comparison_operator=comparison_operator,
                                                                            alert_period=alert_period, topic_id=topic_id, event_id=event_id, where_clause=where_clause)

                # create and activate alert task through the kapacitor HTTP API
                kapacitor_api_tasks_url = "http://{0}:{1}/kapacitor/v1/tasks".format(kapacitor_host, kapacitor_port)
                kapacitor_http_request_body = {
                    "id": alert_id,
                    "template-id": template_id,
                    "dbrps": [{"db": db, "rp": "autogen"}],
                    "status": "enabled",
                    "vars": template_vars
                }

                # send the request and receive a response
                response = post(kapacitor_api_tasks_url, json=kapacitor_http_request_body)
                response_content = response.json()
                # log the response
                log.info(response_content, response.status_code)

                # track all reported errors
                if response_content.get("error", "") != "":
                    alert_tasks_errors.append({
                        "policy": policy_id,
                        "trigger": event_id,
                        "error": response_content.get("error")
                    })

                # extract http handlers
                http_handlers = trigger.trigger_tpl["action"]["implementation"]

                # subscribe all http handlers to the created topic
                self._config_kapacitor_alert_handlers(kapacitor_host, kapacitor_port, sfc, sfc_instance, policy_id, resource_spec_trigger_id, topic_id, event_id,
                                                      http_handlers, alert_handlers_errors)

    def _config_kapacitor_alert_handlers(self, kapacitor_host, kapacitor_port, sfc, sfc_i, policy_id, trigger_id,
                                         topic_id, event_id, http_handlers, alert_handlers_errors):
        """
        Handles the configuration of HTTP Post alert handlers.

        :param kapacitor_host: default host is localhost (CLMC service running on the same machine as Kapacitor)
        :param kapacitor_port: default value to use is 9092
        :param sfc: service function chain identifier
        :param sfc_i: service function chain instance identifier
        :param policy_id: policy ID those triggers relate to
        :param trigger_id: the resource specification trigger ID
        :param topic_id: topic ID built of sfc, sfc instance and event_id
        :param event_id: name of trigger
        :param http_handlers: list of handlers to subscribe
        :param alert_handlers_errors: the list for tracking errors while interacting with Kapacitor alert handlers
        """

        kapacitor_api_handlers_url = "http://{0}:{1}/kapacitor/v1/alerts/topics/{2}/handlers".format(kapacitor_host, kapacitor_port, topic_id)
        for http_handler_url in http_handlers:

            # check for flame_sfemc entry, if found replace with sfemc FQDN
            if http_handler_url == SFEMC:
                sfemc_fqdn = self.request.registry.settings['sfemc_fqdn']
                http_handler_url = "http://{0}:8080/sfemc/event/{1}/{2}/{3}".format(sfemc_fqdn, sfc, policy_id, trigger_id)

            handler_id = "{0}\n{1}\n{2}\n{3}\n{4}".format(sfc, sfc_i, policy_id, event_id, http_handler_url)
            handler_id = self.get_hash(handler_id)
            kapacitor_http_request_body = fill_http_post_handler_vars(handler_id, http_handler_url)
            response = post(kapacitor_api_handlers_url, json=kapacitor_http_request_body)
            response_content = response.json()
            log.info(response_content, response.status_code)

            if response_content.get("error", "") != "":
                alert_handlers_errors.append({
                    "policy": policy_id,
                    "trigger": event_id,
                    "handler": http_handler_url,
                    "error": response_content.get("error")
                })

    @staticmethod
    def get_hash(message):
        """
        Returns the hash value of a message encoded with utf-8 using hash algorithm sha256

        :param message: the message to hash

        :return: the value of the has
        """

        byte_str = bytes(message, encoding="utf-8")
        hash_obj = sha256(byte_str)

        return hash_obj.hexdigest()
